package main

import (
	"flag"
	"fmt"
	"my/go-translator/transpile"
	"my/go-translator/transpile_python"
	"os"
)

const (
	CPP_MODE = iota
	PYTHON_MODE
)

func main() {
	mode, source, target, header_file, addArduinoH := getFlags()
	checkFlagsAreValid(source)
	safeTranspile(mode, source, target, header_file, addArduinoH)
}

func getFlags() (int, string, string, string, bool) {
	pmode := flag.Bool("p", false, "Mode: C++ or Python")
	header_file := flag.String("h", "", "Write headers to file also")
	addArduinoH := flag.Bool("a", false, "Add Arduino.h")
	flag.Parse()

	source := flag.Arg(0)
	target := flag.Arg(1)

	mode := CPP_MODE
	if *pmode {
		mode = PYTHON_MODE
	}

	return mode, source, target, *header_file, *addArduinoH
}

func checkFlagsAreValid(source string) {
	if source == "" {
		printUsage()
		os.Exit(1)
	}
}

func printUsage() {
	fmt.Print("This program transpiles Golang source into corresponding C++/python code.\n\n")
	fmt.Printf("Options:\n")
	flag.PrintDefaults()
	fmt.Print("\n")
	fmt.Print("Example:\n")
	fmt.Printf("\tgo-tr [-a] [-h header.h] [-p] controller.go [output.cpp]\n\n")
}

func safeTranspile(mode int, source, target, header_file string, addArduinoH bool) {
	// Read the Golang source file.
	in, err := os.Open(source)
	if err != nil {
		fmt.Fprintf(os.Stderr, "Go source file [%s] could not be opened! %v", source, err)
		os.Exit(1)
	}
	defer in.Close()

	out := os.Stdout
	if target != "" {
		// Create the Arduino sketch file.
		os.Remove(target)
		out, err = os.OpenFile(target, os.O_CREATE|os.O_RDWR|os.O_SYNC, 0644)
		if err != nil {
			fmt.Fprintf(os.Stderr, "Arduino sketch file [%s] could not be opened! %v", target, err)
			os.Exit(1)
		}
	}

	var header_f *os.File
	if header_file != "" {
		os.Remove(header_file)
		header_f, err = os.OpenFile(header_file, os.O_CREATE|os.O_RDWR|os.O_SYNC, 0644)
		if err != nil {
			fmt.Fprintf(os.Stderr, "Header file [%s] could not be opened! %v", header_file, err)
			os.Exit(1)
		}
	}
	switch mode {
	case CPP_MODE:
		// Transpiles the Golang source into Arduino sketch.
		service := transpile.NewService(in, out)
		service.SetHeaderWriter(header_f)
		service.AddIncludeArduinoH(addArduinoH)
		if err := service.Start(); err != nil {
			fmt.Fprintf(os.Stderr, "%v", err)
			os.Exit(1)
		}
	case PYTHON_MODE:
		service := transpile_python.NewService(in, out)
		if err := service.Start(); err != nil {
			fmt.Fprintf(os.Stderr, "%v", err)
			os.Exit(1)
		}
	}
}
